@@ -1,7 +1,10 @@
 from __future__ import absolute_import
 
+from .logger import configure_logging
+configure_logging()
+
 from .trustar import TruStar
 from .models import *
 from .utils import *
 
-from .version import __version__
+from .version import __version__, __api_version__
