@@ -8,15 +8,16 @@ from six import string_types
 import json
 from datetime import datetime
 import functools
+import logging
 
 # package imports
-from .models import Page, Report, DistributionType, IdType
-from .utils import get_logger, get_time_based_page_generator
+from .models import NumberedPage, Report, RedactedReport, DistributionType, IdType
+from .utils import get_time_based_page_generator, DAY
 
 # python 2 backwards compatibility
 standard_library.install_aliases()
 
-logger = get_logger(__name__)
+logger = logging.getLogger(__name__)
 
 
 class ReportClient(object):
@@ -62,8 +63,8 @@ class ReportClient(object):
         found by adjusting the ``from_time`` and ``to_time`` parameters.
 
         Note:  This endpoint will only return reports from a time window of maximum size of 2 weeks. If you give a
-        time window larger than 2 weeks, it will pull reports starting at 2 weeks before the “to” date, through the
-        “to” date.
+        time window larger than 2 weeks, it will pull reports starting at 2 weeks before the "to" date, through the
+        "to" date.
 
         :param boolean is_enclave: restrict reports to specific distribution type (optional - by default all accessible
             reports are returned).
@@ -75,16 +76,16 @@ class ReportClient(object):
         :param int from_time: start of time window in milliseconds since epoch (optional)
         :param int to_time: end of time window in milliseconds since epoch (optional)
 
-        :return: A |Page| of |Report| objects.
+        :return: A |NumberedPage| of |Report| objects.
 
         """
 
         distribution_type = None
 
         # explicitly compare to True and False to distinguish from None (which is treated as False in a conditional)
-        if is_enclave == True:
+        if is_enclave:
             distribution_type = DistributionType.ENCLAVE
-        elif is_enclave == False:
+        elif not is_enclave:
             distribution_type = DistributionType.COMMUNITY
 
         if enclave_ids is None:
@@ -99,14 +100,14 @@ class ReportClient(object):
             'excludedTags': excluded_tags
         }
         resp = self._client.get("reports", params=params)
-        result = Page.from_dict(resp.json(), content_type=Report)
+        result = NumberedPage.from_dict(resp.json(), content_type=Report)
 
-        # create a Page object from the dict
+        # create a NumberedPage object from the dict
         return result
 
     def submit_report(self, report):
         """
-        Submits a report.
+        Submit a report.
 
         * If ``report.is_enclave`` is ``True``, then the report will be submitted to the enclaves
           identified by ``report.enclaves``; if that field is ``None``, then the enclave IDs registered with this
@@ -145,7 +146,7 @@ class ReportClient(object):
 
         # default time began is current time
         if report.time_began is None:
-            report.time_began = datetime.now()
+            report.set_time_began(datetime.now())
 
         data = json.dumps(report.to_dict())
         resp = self._client.post("reports", data=data, timeout=60)
@@ -219,6 +220,75 @@ class ReportClient(object):
         params = {'idType': id_type}
         self._client.delete("reports/%s" % report_id, params=params)
 
+    def copy_report(self, src_report_id, dest_enclave_id, from_provided_submission=False, report=None, tags=None):
+        """
+        Copy a report to another enclave.  All properties of the report, including tags, will be copied.
+        A reference to the original report will still be stored on the child, allowing the system to track the
+        relationship between the original report and copies made from it.
+
+        If the ``from_provided_submission`` parameter is ``True``, then edits can be applied to the copied report.  This
+        is useful in cases where the body or title must be redacted first, or the list of tags needs to be altered for
+        the copy.  In this case, a |Report| object and a list of tag names must be provided, which will fill out the
+        copied report.  A reference to the original report will still be stored on the copy.
+        **NOTE:** Partial edits are not allowed.  ALL fields must be filled out on this object, and the
+        fields from the original report will completely ignored.
+
+        :param str src_report_id: the ID of the report to copy
+        :param str dest_enclave_id: the ID of the enclave to copy the report to
+        :param boolean from_provided_submission: whether to apply edits from a supplied report object and list of tags
+        :param Report report: (required if ``from_provided_submission`` is ``True``) a report object containing an edited version to use as the copy.
+            This allows information to be redacted, or other arbitrary edits to be made to the copied version.
+            **NOTE:** Partial edits are not allowed.  ALL fields must be filled out on this object, and the fields from
+            the original report will completely ignored.
+        :param list(str) tags: (required if ``from_provided_submission`` is ``True``) a list of tags to use if ``from_provided_submission`` is ``True``.
+            **NOTE:** if ``from_provided_submission`` is True, the tags from the source report will be completely
+            ignored, and this list of tags will be used instead.  MUST be provided if ``from_provided_submission`` is ``True``.
+        :return: the ID of the newly-created copy
+        """
+
+        params = {
+            'destEnclaveId': dest_enclave_id,
+            'copyFromProvidedSubmission': from_provided_submission
+        }
+
+        # determine if edits are being made to the copy
+        if from_provided_submission:
+            # ensure an edited version of the report has been provided
+            if not report:
+                raise Exception("Cannot copy from provided submission without providing a report object")
+            # ensure an edited list of tags has been provided
+            if not tags:
+                raise Exception("Cannot copy from provided submission without providing a list of tags")
+
+            # form the JSON dictionary of the report
+            body = report.to_dict()
+            # add the list of tags to the JSON
+            # NOTE: this field on the report object cannot be used in other endpoints on this API version
+            body['tags'] = tags
+        else:
+            body = None
+
+        response = self._client.post('reports/copy/{id}'.format(id=src_report_id), params=params, data=json.dumps(body))
+        return response.json().get('id')
+
+    def move_report(self, report_id, dest_enclave_id):
+        """
+        Move a report from one enclave to another.
+
+        **NOTE:** All tags will be moved, as well.
+
+        :param report_id: the ID of the report to move
+        :param dest_enclave_id: the ID of the enclave to move the report to
+        :return: the ID of the report
+        """
+
+        params = {
+            'destEnclaveId': dest_enclave_id
+        }
+
+        response = self._client.post('/reports/move/{id}'.format(id=report_id), params=params)
+        return response.json().get('id')
+
     def get_correlated_report_ids(self, indicators):
         """
         DEPRECATED!
@@ -246,6 +316,8 @@ class ReportClient(object):
         :param indicators: A list of indicator values to retrieve correlated reports for.
         :param enclave_ids: The enclaves to search in.
         :param is_enclave: Whether to search enclave reports or community reports.
+        :param int page_number: the page number to get.
+        :param int page_size: the size of the page to be returned.
         :return: The list of IDs of reports that correlated.
 
         Example:
@@ -263,33 +335,55 @@ class ReportClient(object):
         params = {
             'indicators': indicators,
             'enclaveIds': enclave_ids,
-            'distributionType': distribution_type
+            'distributionType': distribution_type,
+            'pageNumber': page_number,
+            'pageSize': page_size
         }
         resp = self._client.get("reports/correlated", params=params)
 
-        return Page.from_dict(resp.json(), content_type=Report)
+        return NumberedPage.from_dict(resp.json(), content_type=Report)
 
-    def search_reports_page(self, search_term, enclave_ids=None, page_size=None, page_number=None):
+    def search_reports_page(self, search_term=None,
+                            enclave_ids=None,
+                            from_time=None,
+                            to_time=None,
+                            tags=None,
+                            excluded_tags=None,
+                            page_size=None,
+                            page_number=None):
         """
         Search for reports containing a search term.
 
-        :param str search_term: The term to search for.
+        :param str search_term: The term to search for.  If empty, no search term will be applied.  Otherwise, must
+            be at least 3 characters.
         :param list(str) enclave_ids: list of enclave ids used to restrict reports to specific enclaves (optional - by
             default reports from all of user's enclaves are returned)
-        :param int page_number: the page number to get.
+        :param int from_time: start of time window in milliseconds since epoch (optional)
+        :param int to_time: end of time window in milliseconds since epoch (optional)
+        :param list(str) tags: Name (or list of names) of tag(s) to filter reports by.  Only reports containing
+            ALL of these tags will be returned. (optional)
+        :param list(str) excluded_tags: Reports containing ANY of these tags will be excluded from the results.
+        :param int page_number: the page number to get. (optional)
         :param int page_size: the size of the page to be returned.
-        :return: a |Page| of |Report| objects.
+        :return: a |NumberedPage| of |Report| objects.  *NOTE*:  The bodies of these reports will be ``None``.
         """
 
+        body = {
+            'searchTerm': search_term
+        }
+
         params = {
-            'searchTerm': search_term,
             'enclaveIds': enclave_ids,
+            'from': from_time,
+            'to': to_time,
+            'tags': tags,
+            'excludedTags': excluded_tags,
             'pageSize': page_size,
             'pageNumber': page_number
         }
 
-        resp = self._client.get("reports/search", params=params)
-        page = Page.from_dict(resp.json(), content_type=Report)
+        resp = self._client.post("reports/search", params=params, data=json.dumps(body))
+        page = NumberedPage.from_dict(resp.json(), content_type=Report)
 
         return page
 
@@ -310,29 +404,52 @@ class ReportClient(object):
         :return: The generator.
         """
 
+        def get_next_to_time(result, to_time):
+            """
+            For each page, get the timestamp of the earliest report in the result set.  The next query will use this
+            timestamp as the end of its interval.  This endpoint limits queries to 1 day.  If the result set is
+            empty, subtract 1 day from the to_time for the next interval.
+
+            :param result: the result set of the previous call
+            :param to_time: the to_time of the previous call
+            :return: the next to_time
+            """
+
+            if len(result.items) > 0:
+                return result.items[-1].updated - 1
+            else:
+                return to_time - DAY
+
         get_page = functools.partial(self.get_reports_page, is_enclave, enclave_ids, tag, excluded_tags)
         return get_time_based_page_generator(
             get_page=get_page,
-            get_next_to_time=lambda x: x.items[-1].updated if len(x.items) > 0 else None,
+            get_next_to_time=get_next_to_time,
             from_time=from_time,
             to_time=to_time
         )
 
     def get_reports(self, is_enclave=None, enclave_ids=None, tag=None, excluded_tags=None, from_time=None, to_time=None):
         """
-        Uses the |get_reports_page| method to create a generator that returns each successive report.
+        Uses the |get_reports_page| method to create a generator that returns each successive report as a trustar
+        report object.
 
         :param boolean is_enclave: restrict reports to specific distribution type (optional - by default all accessible
             reports are returned).
         :param list(str) enclave_ids: list of enclave ids used to restrict reports to specific
             enclaves (optional - by default reports from all enclaves are returned)
-        :param str tag: name of tag to filter reports by.  if a tag with this name exists in more than one enclave
-            indicated in ``enclave_ids``, the request will fail.  handle this by making separate requests for each
+        :param list(str) tag: a list of tags; only reports containing ALL of these tags will be returned. 
+            If a tag with this name exists in more than one enclave in the list passed as the ``enclave_ids``
+            argument, the request will fail.  Handle this by making separate requests for each
             enclave ID if necessary.
+        :param list(str) excluded_tags: a list of tags; reports containing ANY of these tags will not be returned. 
         :param int from_time: start of time window in milliseconds since epoch (optional)
         :param int to_time: end of time window in milliseconds since epoch (optional)
-        :return: The generator.
+        :return: A generator of Report objects.
 
+        Note:  If a report contains all of the tags in the list passed as argument to the 'tag' parameter and also 
+        contains any (1 or more) of the tags in the list passed as argument to the 'excluded_tags' parameter, that 
+        report will not be returned by this function.  
+        
         Example:
 
         >>> page = ts.get_reports(is_enclave=True, tag="malicious", from_time=1425695711000, to_time=1514185311000)
@@ -345,9 +462,9 @@ class ReportClient(object):
 
         """
 
-        return Page.get_generator(page_generator=self._get_reports_page_generator(is_enclave, enclave_ids, tag,
+        return NumberedPage.get_generator(page_generator=self._get_reports_page_generator(is_enclave, enclave_ids, tag,
                                                                                   excluded_tags, from_time, to_time))
-    
+
     def _get_correlated_reports_page_generator(self, indicators, enclave_ids=None, is_enclave=True,
                                                start_page=0, page_size=None):
         """
@@ -361,7 +478,7 @@ class ReportClient(object):
         """
 
         get_page = functools.partial(self.get_correlated_reports_page, indicators, enclave_ids, is_enclave)
-        return Page.get_page_generator(get_page, start_page, page_size)
+        return NumberedPage.get_page_generator(get_page, start_page, page_size)
 
     def get_correlated_reports(self, indicators, enclave_ids=None, is_enclave=True):
         """
@@ -373,33 +490,78 @@ class ReportClient(object):
         :return: The generator.
         """
 
-        return Page.get_generator(page_generator=self._get_correlated_reports_page_generator(indicators,
+        return NumberedPage.get_generator(page_generator=self._get_correlated_reports_page_generator(indicators,
                                                                                              enclave_ids,
                                                                                              is_enclave))
-    
-    def _search_reports_page_generator(self, search_term, enclave_ids=None, start_page=0, page_size=None):
+
+    def _search_reports_page_generator(self, search_term=None,
+                                       enclave_ids=None,
+                                       from_time=None,
+                                       to_time=None,
+                                       tags=None,
+                                       excluded_tags=None,
+                                       start_page=0,
+                                       page_size=None):
         """
         Creates a generator from the |search_reports_page| method that returns each successive page.
 
-        :param str search_term: The term to search for.
+        :param str search_term: The term to search for.  If empty, no search term will be applied.  Otherwise, must
+            be at least 3 characters.
         :param list(str) enclave_ids: list of enclave ids used to restrict reports to specific enclaves (optional - by
             default reports from all of user's enclaves are returned)
+        :param int from_time: start of time window in milliseconds since epoch (optional)
+        :param int to_time: end of time window in milliseconds since epoch (optional)
+        :param list(str) tags: Name (or list of names) of tag(s) to filter reports by.  Only reports containing
+            ALL of these tags will be returned. (optional)
+        :param list(str) excluded_tags: Reports containing ANY of these tags will be excluded from the results.
         :param int start_page: The page to start on.
         :param page_size: The size of each page.
         :return: The generator.
         """
 
-        get_page = functools.partial(self.search_reports_page, search_term, enclave_ids)
-        return Page.get_page_generator(get_page, start_page, page_size)
+        get_page = functools.partial(self.search_reports_page, search_term, enclave_ids, from_time, to_time, tags,
+                                     excluded_tags)
+        return NumberedPage.get_page_generator(get_page, start_page, page_size)
 
-    def search_reports(self, search_term, enclave_ids=None):
+    def search_reports(self, search_term=None,
+                       enclave_ids=None,
+                       from_time=None,
+                       to_time=None,
+                       tags=None,
+                       excluded_tags=None):
         """
         Uses the |search_reports_page| method to create a generator that returns each successive report.
 
-        :param str search_term: The term to search for.
+        :param str search_term: The term to search for.  If empty, no search term will be applied.  Otherwise, must
+            be at least 3 characters.
         :param list(str) enclave_ids: list of enclave ids used to restrict reports to specific enclaves (optional - by
             default reports from all of user's enclaves are returned)
-        :return: The generator.
+        :param int from_time: start of time window in milliseconds since epoch (optional)
+        :param int to_time: end of time window in milliseconds since epoch (optional)
+        :param list(str) tags: Name (or list of names) of tag(s) to filter reports by.  Only reports containing
+            ALL of these tags will be returned. (optional)
+        :param list(str) excluded_tags: Reports containing ANY of these tags will be excluded from the results.
+        :return: The generator of Report objects.  Note that the body attributes of these reports will be ``None``.
         """
 
-        return Page.get_generator(page_generator=self._search_reports_page_generator(search_term, enclave_ids))
+        return NumberedPage.get_generator(page_generator=self._search_reports_page_generator(search_term, enclave_ids,
+                                                                                     from_time, to_time, tags,
+                                                                                     excluded_tags))
+
+    def redact_report(self, title=None, report_body=None):
+        """
+        Redacts a report's title and body.
+
+        :param str title: The title of the report to apply redaction to.
+        :param str report_body: The body of the report to apply redaction to.
+        :return: a |RedactedReport| object.
+        """
+
+        body = {
+            'title': title,
+            'reportBody': report_body
+        }
+
+        resp = self._client.post("redaction/report", data=json.dumps(body))
+
+        return RedactedReport.from_dict(resp.json())
